#include "pyggi_visual_input.inc" 
#include "pyggi_visual_drawing.inc" 
#include "pyggi_visual_blit.inc" 
#include "pyggi_pack.inc" 

static char vis_flush_doc[]="
Flush the visual.

Parameter: ()|(i,i,i,i) : the area to flush (absolute coords) (all if not set)
Return:    None
";
static PyObject* vis_flush(GGIVisualObject*self, PyObject*args) {
	int x1,y1,x2,y2;
	
	if(PyTuple_GET_SIZE(args)==0) 
		ggiFlush(self->vis);
	else {
		if(!PyArg_ParseTuple(args, "iiii", &x1, &y1, &x2, &y2)) {
			return NULL;
		}
		ggiFlushRegion(self->vis, x1,y1, (x2-x1+1), (y2-y1+1));
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static char vis_fillscreen_doc[]="
Fill the entire screen.

Parameters : ()|(i)|([f,f,f]) The fill color (fg if not set)
Return:      None
";
static PyObject* vis_fillscreen(GGIVisualObject*self, PyObject*args) {
	ggi_pixel p,fg;
	
	switch(PyTuple_GET_SIZE(args)) {
	case 0:
		ggiFillscreen(self->vis);
		Py_INCREF(Py_None);
		return Py_None;
	case 1:
		if(pyggi_get_pixel(PyTuple_GET_ITEM(args,0), &p, self->vis)<0) {
			return NULL;
		}
		ggiGetGCForeground(self->vis, &fg);
		ggiSetGCForeground(self->vis, p);
		ggiFillscreen(self->vis);
		ggiSetGCForeground(self->vis, fg);
		Py_INCREF(Py_None);
		return Py_None;
		
	default:
	}
	PyErr_SetString(PyExc_TypeError, "Need at most one argument");
	return NULL;
}



/****************************************************************/

enum {
	PYGGI_METH_PIXEL,
	PYGGI_METH_LINE,
	PYGGI_METH_HLINE,
	PYGGI_METH_VLINE,
	PYGGI_METH_BOX,
	
	PYGGI_METH_CROSSBLIT,
	PYGGI_METH_COPYBOX,
	
	PYGGI_METH_PACK,
	
	PYGGI_METH_MODE,
	PYGGI_METH_FILLSCREEN,
	PYGGI_METH_FLUSH,

	PYGGI_METH_JOIN,
	PYGGI_METH_DETACH,
	PYGGI_METH_READ,
	PYGGI_METH_POLL,
	PYGGI_METH_BLOCK,
	PYGGI_METH_QUEUED,
	PYGGI_METH_SEND,
};

static struct PyMethodDef vis_methods[] = {
	
	/* ---------- drawing ops --------------------*/
	{ "pixel", (PyCFunction)vis_pixel, METH_VARARGS, (char*)vis_pixel_doc},
	{ "line",  (PyCFunction)vis_line,  METH_VARARGS, (char*)vis_line_doc},
	{ "hline", (PyCFunction)vis_hline, METH_VARARGS, (char*)vis_hline_doc},
	{ "vline", (PyCFunction)vis_vline, METH_VARARGS, (char*)vis_vline_doc},
	{ "box",   (PyCFunction)vis_box,   METH_VARARGS, (char*)vis_box_doc},

	/* ----------- blitting ops -----------------*/

	{ "crossblit", (PyCFunction)vis_crossblit, METH_VARARGS, (char*)vis_crossblit_doc},
	{ "copybox",   (PyCFunction)vis_copybox,   METH_VARARGS, (char*)vis_copybox_doc},

	{ "Pack",   (PyCFunction)PackConstruct,   METH_VARARGS, (char*)pack_construct_doc},

	/* ---------- visual ops --------------------*/
	{ "Mode", (PyCFunction)ModeConstruct,  METH_VARARGS, (char*)mode_construct_doc},
	{ "fill", (PyCFunction)vis_fillscreen, METH_VARARGS, (char*)vis_fillscreen_doc},
	{ "flush",(PyCFunction)vis_flush,      METH_VARARGS, (char*)vis_flush_doc },
	
	/* ---------- Input management ----------------*/
	{ "join",    (PyCFunction)vis_join,    METH_VARARGS, (char*)vis_join_doc  },
	{ "detach",  (PyCFunction)vis_detach,  METH_NOARGS,  (char*)vis_detach_doc },

	/*------------- Events functions ------------------------*/
	{ "read",  (PyCFunction)vis_eventread,    METH_VARARGS, (char*)vis_eventread_doc },
	{ "poll",  (PyCFunction)vis_eventpoll,    METH_VARARGS, (char*)vis_eventpoll_doc },
	{ "wait",  (PyCFunction)vis_eventblock,   METH_VARARGS, (char*)vis_eventblock_doc },
	{ "queued",(PyCFunction)vis_eventsqueued, METH_VARARGS, (char*)vis_eventsqueued_doc},
	{ "send",  (PyCFunction)vis_eventsend,    METH_VARARGS, (char*)vis_eventsend_doc},

	{NULL, NULL}
};

static PyObject * vis_default_getattr(GGIVisualObject *self, char*name) {
	ggi_pixel p;
	ggi_mode mode;
	int a,b,c,d;
	PyObject *r;
	
	switch(name[0]) {		
	case 'b':
		if(!strcmp(name,"box"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_BOX],(PyObject*)self);
		if(!strcmp(name,"bg")) {
			ggiGetGCBackground(self->vis,&p);
			return PyInt_FromLong(p);
		}
		break;
	case 'c':
		if(!strcmp(name,"clipping")) {
			ggiGetGCClipping(self->vis,&a,&b,&c,&d);
			r=PyList_New(4);
			PyList_SET_ITEM(r,0,PyInt_FromLong(a));
			PyList_SET_ITEM(r,1,PyInt_FromLong(b));
			PyList_SET_ITEM(r,2,PyInt_FromLong(c));
			PyList_SET_ITEM(r,3,PyInt_FromLong(d));
			return r;
		}
		if(!strcmp(name,"copybox"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_COPYBOX],(PyObject*)self);
		break;
	case 'd':
		if(!strcmp(name,"displayframe"))
			return PyInt_FromLong(ggiGetDisplayFrame(self->vis));
		if(!strcmp(name,"devices"))
			return gii_c_api->deviceinfo(self->vis->input);
		if(!strcmp(name,"detach"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_DETACH],(PyObject*)self);
		break;
	case 'f':
		if(!strcmp(name,"flush"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_FLUSH],(PyObject*)self);
		if(!strcmp(name,"fill"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_FILLSCREEN],(PyObject*)self);
		if(!strcmp(name,"flags"))
			return PyInt_FromLong(ggiGetFlags(self->vis));
		if(!strcmp(name,"fg")) {
			ggiGetGCForeground(self->vis,&p);
			return PyInt_FromLong(p);
		}
		break;
	case 'g': /* gamma */
		break;
	case 'h':
		if(!strcmp(name,"hline"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_HLINE],(PyObject*)self);
		break;
	case 'j':
		if(!strcmp(name,"join"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_JOIN],(PyObject*)self);
		break;
	case 'l':
		if(!strcmp(name,"line"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_LINE],(PyObject*)self);
		break;
	case 'm':
		if(!strcmp(name,"mask")) {
			return PyInt_FromLong(ggiGetEventMask(self->vis));
		}
		if(!strcmp(name,"mode")) {
			ggiGetMode(self->vis,&mode);
			return newModeFromGGIMode(mode);
		}		
		break;
	case 'o':
		if(!strcmp(name,"origin")) {
			ggiGetOrigin(self->vis,&a,&b);
			r=PyList_New(2);
			PyList_SET_ITEM(r,0,PyInt_FromLong(a));
			PyList_SET_ITEM(r,1,PyInt_FromLong(b));
			return r;
		}
		break;
	case 'p': /* put(), palette */
		if(!strcmp(name,"pixel"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_PIXEL],(PyObject*)self);
		if(!strcmp(name,"poll"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_POLL],(PyObject*)self);
		break;
	case 'q':
		if(!strcmp(name,"queued"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_QUEUED],(PyObject*)self);
		break;
	case 'r':
		if(!strcmp(name,"read"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_READ],(PyObject*)self);
		if(!strcmp(name,"readframe"))
			return PyInt_FromLong(ggiGetReadFrame(self->vis));
		break;
	case 's':
		if(!strcmp(name,"send"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_SEND],(PyObject*)self);
		break;
	case 'v':
		if(!strcmp(name,"vline"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_VLINE],(PyObject*)self);
		break;
	case 'w':
		if(!strcmp(name,"wait"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_BLOCK],(PyObject*)self);
		if(!strcmp(name,"writeframe"))
			return PyInt_FromLong(ggiGetWriteFrame(self->vis));
		break;
	case 'M':
		if(!strcmp(name,"Mode"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_MODE],(PyObject*)self);
		break;
	case 'P':
		if(!strcmp(name,"Pack"))
			return PyCFunction_New(&vis_methods[PYGGI_METH_PACK],(PyObject*)self);
		break;
		
	}
	/* Not found */
	return NULL;
}

/*
  return 0 if not found 1 if ok -1 on error
 */
static int vis_default_setattr(GGIVisualObject *self, char*name, PyObject * value) {
	int i;
	ggi_pixel p;
	ggi_coord c;
	int r[4];

	switch(name[0]) {
	case 'b':
		if(!strcmp(name,"bg")) {
			if(pyggi_get_pixel(value, &p, self->vis)<0)
				return -1;
			ggiSetGCBackground(self->vis,p);
			return 1;
		}
		break;
	case 'c':
		if(!strcmp(name,"clipping")) {
			if(pyggi_get_rect(value, r)<0)
				return -1;
			if(ggiSetGCClipping(self->vis,r[0],r[1],r[2],r[3])) {
				PyErr_SetString(PyExc_SystemError, "ggiSetGCClipping failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'd':
		if(!strcmp(name,"displayframe")) {
			if(pyggi_get_int(value, &i)<0)
				return -1;
			if(ggiSetDisplayFrame(self->vis,i)) {
				PyErr_SetString(PyExc_SystemError, "ggiSetDisplayFrame failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'f':
		if(!strcmp(name,"fg")) {
			if(pyggi_get_pixel(value, &p,self->vis)<0)
				return -1;
			ggiSetGCForeground(self->vis,p);
			return 1;
		}
		if(!strcmp(name,"flags")) {
			if(pyggi_get_int(value, &i)<0)
				return -1;
			ggiSetFlags(self->vis, i);
			return 1;
		}
		break;
	case 'g': /* gamma */
		break;
	case 'm':
		if(!strcmp(name,"mask")) {
			if(pyggi_get_int(value, &i)<0)
				return -1;
			ggiSetEventMask(self->vis, i);
			return 1;
		}
		if(!strcmp(name,"mode")) {
			if(!GGIMode_Check(value)) {
				PyErr_SetString(PyExc_TypeError, "Mode required");
				return -1;
			}
			if(ggiSetMode(self->vis, &((GGIModeObject*)value)->mode)) {
				PyErr_SetString(PyExc_SystemError, "ggiSetMode failed");
				return -1;
			}
			ggiGetMode(self->vis,&(self->mode));
			self->pf=ggiGetPixelFormat(self->vis);
			return 1;
		}
		break;
	case 'o':
		if(!strcmp(name,"origin")) {
			if(pyggi_get_coords(value, &c)<0)
				return -1;
			if(ggiSetOrigin(self->vis,c.x,c.y)) {
				PyErr_SetString(PyExc_SystemError, "ggiSetOrigin failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'p': /* palette */
		break;
	case 'r':
		if(!strcmp(name,"readframe")) {
			if(pyggi_get_int(value, &i)<0)
				return -1;
			if(ggiSetReadFrame(self->vis,i)) {
				PyErr_SetString(PyExc_SystemError, "ggiSetReadFrame failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'w':
		if(!strcmp(name,"writeframe")) {
			if(pyggi_get_int(value, &i)<0)
				return -1;
			if(ggiSetWriteFrame(self->vis,i)) {
				PyErr_SetString(PyExc_SystemError, "ggiSetWriteFrame failed");
				return -1;
			}
			return 1;
		}
		break;
	}
	/* Not found */
	return 0;
}


/* */
static void vis_dealloc(GGIVisualObject *self)  {
	pyggi_extension_detach_all(self);
	ggiClose(self->vis);
	PyObject_DEL(self);
}

static PyObject* vis_getattr(GGIVisualObject *self, char*name)  {
	GGIExtension *e;
	PyObject * r;

	/* Priority handlers  */	
	if(name[0]=='e' && !strcmp(name,"extensions")) {
		return pyggi_extension_list(self);
	}
	
	
	/* Search in extensions */
	for(e=self->exts;e;e=e->next) {
		r=e->getattr(self,name);
		if(r || PyErr_Occurred())
			return r;
	}
	
	/* Fallback to default handler */
	r=vis_default_getattr(self,name);
	if(r || PyErr_Occurred())
		return r;
	
	/* Not found */

	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static int vis_setattr(GGIVisualObject *self,
		       char* name,
		       PyObject * v) {
	GGIExtension *e;
	int r;
	
	/* Priority handlers  */
	
	/* Search in extensions */
	for(e=self->exts;e;e=e->next) {
		r=e->setattr(self,name,v);
		if(r==1)
			return 0;
		else if(r==-1)
			return -1;
	}
	
	/* Fallback to default handler */
	
	r=vis_default_setattr(self,name,v);
	if(r==1)
		return 0;
	else if(r==-1)
		return -1;
	
	/* Not found */	
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

static PyTypeObject vis_type = {
	PyObject_HEAD_INIT(NULL)
	0,
	"GGI Visual",
	sizeof(GGIVisualObject),
	0,
	(destructor)vis_dealloc,
	(printfunc)0,
	(getattrfunc)vis_getattr,
	(setattrfunc)vis_setattr,
	(cmpfunc)0,
	(reprfunc)0,
	0,
	0,
	0,
	(hashfunc)0,
	(ternaryfunc)0,
	(reprfunc)0,
	
	0L,0L,0L,0L,
	"no docs."
};

// begin-bad-hack
typedef ggi_visual_t _varidx(char**);
typedef _varidx*varidx;
#define v_ ggi_visual_t
v_ open0(char**v){return ggiOpen(NULL);} 
v_ open1(char**v){return ggiOpen(v[0],NULL);} 
v_ open2(char**v){return ggiOpen(v[0],v[1],NULL);} 
v_ open3(char**v){return ggiOpen(v[0],v[1],v[2],NULL);} 
v_ open4(char**v){return ggiOpen(v[0],v[1],v[2],v[3], NULL);} 
v_ open5(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],NULL);} 
v_ open6(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],NULL);} 
v_ open7(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],NULL);} 
v_ open8(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],NULL);}
v_ open9(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],NULL);}
v_ open10(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],NULL);}

#define kcah(n) &open ## n
varidx openWithVisuals[] ={kcah(0),kcah(1),kcah(2),kcah(3),kcah(4),kcah(5),
			   kcah(6),kcah(7),kcah(8),kcah(9),kcah(10)};
#undef v_
#undef kcah
// bad-hacks-never-end
static PyObject* newVisualObject(PyObject *targets) {
	GGIVisualObject *self;
	char *visuals[10];
	PyObject *t;
	int i,size;
	
	self=PyObject_New(GGIVisualObject, &vis_type);
	if(!self) return NULL;
	
	size=PyTuple_Size(targets);
	if(size>10) {
		PyErr_SetString(PyExc_TypeError,"can only manage 10 arguments.");
		return NULL;
	}
	
	for(i=0;i<size; i++) {
		t=PyTuple_GetItem(targets, i);
		if(!PyString_Check(t)) {
			PyErr_SetString(PyExc_TypeError, "strings only.");
			return NULL;
		}
		visuals[i]=PyString_AsString(t);
	}
	
	
	self->vis=openWithVisuals[size](visuals);
	if(self->vis==NULL) {
		PyErr_SetString(PyExc_SystemError, "ggiOpen returned Null!");
		return NULL;
	}

	ggiGetMode(self->vis,&(self->mode));
	/* TODO : Doesn't it depends on the mode ? */
	self->pf=ggiGetPixelFormat(self->vis);	

	self->exts=NULL;
	self->nb_exts=0;
	return (PyObject*)self;
}
